/*
 * GD.C
 * 1991-12-13
 *
 * Lee Daniel Crocker           lee@mport.com (uunet!mport!lee)
 * 5506 Camden Ave #D3          73407.2030@compuserve.com
 * San Jose, CA  95124
 *
 * A program to list GIFs in a directory with info.
 *
 * This code is released into the public domain completely free of
 * any copyrights and restrictions.  You are welcome (indeed
 * encouraged) to use, modify, and spread this code throughout the
 * known universe either freely or commercially.  Copy it, put your
 * name in the credits, and give it to your friends.  Put bugs in
 * it and give it to your enemies.  I don't care.  If you make
 * useful modifications, you are encouraged to send them to me for
 * future enhancement.
 *
 * Please note: this code is meant to be portable among most systems
 * which conform to ANSI X3.159-1989 with a few extensions not
 * provided by that standard.  In particular, a lot of the code may
 * like it does a lot of work for nothing to avoid assumptions about
 * the underlying character set.  Please think twice about changing
 * this code in the name of "simplification" or "optimization".
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>

#include "pdirect.h"

/*
 * OS dependencies
 */

#define MAX_COLUMNS 256
#define BANG 0x21

#ifdef MSDOS
# define LFCHAR '\xFE'  /* In comments, display newlines as blocks */
#else
# define LFCHAR '~'
#endif

#ifdef MSDOS            /* Need non-ANSI getcwd() */
# ifdef __TURBOC__
#  include <dir.h>
# else
#  include <direct.h>
# endif
#elif defined(UNIX)
# include <unistd.h>
#else
char *
getcwd(char *buf, int len)
{
    if (len < 2) return NULL;
    else {
        strcpy(buf, ".");
        return buf;
    }
}
#endif

/*
 * Prototypes
 */

static char *basename(char *);
static void get_environment(char *);
static int usage(int);
static void fatal(int, char *, ...);
static void show_count(void);
static void show_appblock(char *);
static void show_comments(char *, int, int);
static void print_gif_data(DirEntry *);
static void showfile(Directory *, DirEntry *);
static void listdir(char *);

#define FATAL1(x) fatal(__LINE__,(x))
#define FATAL2(x,y) fatal(__LINE__,(x),(y))
#define FATAL3(x,y,z) fatal(__LINE__,(x),(y),(z))

/*
 * Globals
 */

static int verbose = 0, wide_format = 0;
static int file_count, columns, start_column = 0;
static char command_name[FILENAME_MAX];
static unsigned char iobuf[1024];
static size_t bufcount;

/*
 * Strings
 */

static char Sgif[4] = { 0x47, 0x49, 0x46, 0x00 };

/*
 * Code begins here
 */

int
main(int argc, char *argv[])
{
    int c;
    static int dir_count = 0;
    char current_dir[FILENAME_MAX];

    get_environment(argv[0]);

    while (--argc) {
        if (**++argv == '-') {      /* Command switch */
            while ((c = *++*argv) != '\0') {
                switch (tolower(c)) {
                case 'v':
                    verbose = 1;
                    break;
                case 'w':
                    wide_format = 1;
                    break;
                case 'h':
                case '?':
                    return usage(1);
                default:
                    return usage(0);
                }
            }
        } else {                    /* Directory name on command line */
            ++dir_count;
            listdir(*argv);
        }
    }
    if (dir_count == 0) {
        if (getcwd(current_dir, FILENAME_MAX) != NULL) {
            listdir(current_dir);
        }
    }
    return EXIT_SUCCESS;
}

/*
 * Return a pointer to the final element in a pathname.
 */

#ifdef MSDOS
static char pathseps[] = ":/\\";
#else
static char pathseps[] = "/";
#endif

static char *
basename(char *pathname)
{
    int i, len;
    char *cp, *rcp = pathname;

    len = (int)strlen(pathseps);
    for (i = 0; i < len; ++i) {
        if ((cp = strrchr(rcp, pathseps[i])) != NULL) rcp = cp+1;
    }
    return rcp;
}

/*
 * Get command name for Usage note.  This strips any path that may
 * have preceded the command, and strips and extension on MSDOS.
 * Pre-3.0 versions of MSDOS return the string "C" in argv[0].
 * Also gets stuff from the environment needed by this program.
 */

#ifdef MSDOS
#include <dos.h>
static union REGS regs;
#endif

static void
get_environment(char *arg0)
{
    char *cp;

#ifdef MSDOS
    if (strcmp(arg0, "C") == 0) {
        strcpy(command_name, "GD");
    } else
#endif
    strcpy(command_name, basename(arg0));
#ifdef MSDOS
    if ((cp = strrchr(command_name, '.')) != NULL) {
        *cp = '\0';
    }
#endif

#if defined(MSDOS)
    regs.h.ah = 0x0F;
    int86(0x10, &regs, &regs);
    columns = (int)(regs.h.ah);
#elif defined(UNIX)
    if ((cp = getenv("COLUMNS")) == NULL) {
        columns = 80;
    } else {
        columns = atoi(cp);
        if (columns > MAX_COLUMNS) columns = MAX_COLUMNS;
    }
#else
    columns = 80;
#endif
}

/*
 * Print usage note and optionally banner.  Banner is printed if
 * GC is invoked with no arguments or with -h or -?.
 */

static int
usage(int banner)
{
    if (banner) {
        fprintf(stderr, "%s: GIF directory program\n"
                        "Version 1.00 1991-12-13\n"
                        "by Lee Daniel Crocker\n"
                        "73407.2030@compuserve.com\n\n", command_name);
    }
    fprintf(stderr, "Usage:\n"
                    "%s [-h?] [directory ...]\n\n", command_name);
    fprintf(stderr, "-h or -? prints this usage note.\n\n");

    return EXIT_FAILURE;
}

/*
 * Print fatal error message and exit.
 */

static void
fatal(int lineno, char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    fprintf(stderr, "Error (line %d): ", lineno);
    vfprintf(stderr, fmt, args);
    va_end(args);

    exit(EXIT_FAILURE);
}

/*
 * Tell the user what we've accomplished (with correct grammar).
 */

static void
show_count(void)
{
    char temp[12];

    if (verbose) {
        sprintf(temp, "%d", file_count);
        printf("\n%s file%sfound.\n",
            ((file_count == 0) ? "No" : temp),
            ((file_count == 1) ? " " : "s "));
    }
}

/*
 * Print the text of the comment from the input buffer.
 *
 * This one takes a lot of explaining.  We only have a few columns to
 * work with, and comments may be long and have lots of whitespace, so
 * this routine goes to a lot of effort to compress comment text.  It
 * also assumes that the comments in a GIF are ASCII even if the host
 * is not (although it allows one non-ASCII barbarism: CR-delimited
 * lines as from Mac text files).
 */

static char glyphs[] =  " !\"#$%&'()*+,-./0123456789:;<=>?"
                        "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
                        "`abcdefghijklmnopqrstuvwxyz{|}~";

static void
show_comments(char *buf, int count, int max_chars)
{
    int c, blockc, state;
    char comment[MAX_COLUMNS+4];
    char *bp, *cp, *bplimit, *cplimit;

    cp = comment;
    bp = buf;
    blockc = state = 0;
    cplimit = comment + max_chars;
    bplimit = buf + count;

    for (;;) {
        if (blockc == 0) {      /* Start new sub-block */
            blockc = *bp++;
            if (blockc == 0) break;
        }
        if (bp >= bplimit || cp >= cplimit) break;

        c = *bp++;
        --blockc;

        if (c == 0x09) c = 0x20;
        if (c == 0x0C || c == 0x0D) c = 0x0A;
        if ((c < 0x20 || c > 0x7E) && c != 0x0A) continue;

        switch (state) {
        case 0:             /* State 0: wait for printable */
            if (c <= 0x20) break;
            else state = 1;
            /* fall */
        case 1:
            if (c == 0x0A) {
                state = 3;
            } else {
                *cp++ = glyphs[c - 0x20];
                if (c == 0x20) state = 2;
            }
            break;
        case 2:             /* State 2: compressing spaces */
            if (c == 0x20) break;

            if (c == 0x0A) {
                state = 3;
            } else {
                *cp++ = glyphs[c - 0x20];
                state = 1;
            }
            break;
        case 3:             /* State 3: pending NL */
            if (c > 0x20) {
                *cp++ = LFCHAR;
                *cp++ = glyphs[c - 0x20];
                state = 1;
            }
            break;
        default:
            break;
        }
    }
    *cplimit = '\0';    /* In case we ran over */
    *cp = '\0';
    printf("%s", comment);
}

static void
show_appblock(char *buf)
{
    int i, c;
    char appid[12];

    appid[0] = '[';
    appid[9] = ']';
    appid[10] = '\0';

    for (i = 1; i <= 8; ++i) {
        c = *++buf;
        if (c < 0x20 || c > 0x7E) {
            appid[i] = '?';
        } else {
            appid[i] = glyphs[c - 0x20];
        }
    }
    printf("%s", appid);
}

/*
 * We've found the file, now list its info and call show_comments().
 */

static void
print_gif_data(DirEntry *dep)
{
    char sig[4];
    unsigned width, height, colors, crez, offset;
    int id, char_count;
    static char *crvals[] = { "8","64","512","4k","32k","256k","2M","16M" };

    strncpy(sig, (char*)iobuf+3, 3);
    sig[3] = '\0';

    width = iobuf[6] + (iobuf[7] << 8);
    height = iobuf[8] + (iobuf[9] << 8);
    colors = 1 << ((iobuf[10] & 7) + 1);
    crez = ((iobuf[10] >> 4) & 0x07);

    /*
     * Somebody actually using the return value from printf!
     */
    char_count = printf("%-12.12s%4ux%4ux%4u/%4s%4ldk ", dep->d_name,
        width, height, colors, crvals[crez], (dep->d_size >> 10));

    if (wide_format) {
        start_column += char_count + 2;
        if ((start_column + char_count) < columns) {
            printf("  ");
        } else {
            putchar('\n');
            start_column = 0;
        }
    } else {
        offset = 13 + 3 * colors;
        if ((offset + 2) < bufcount && iobuf[offset] == BANG) {
            id = iobuf[++offset];
            ++offset;
        
            if (id == 0xFE) {
                show_comments((char*)iobuf + offset, bufcount - offset,
                    columns - (char_count + 1));
            } else if (id == 0xFF) {
                show_appblock((char*)iobuf + offset);
            }
        }
        putchar('\n');
    }
    ++file_count;
}


/*
 *
 */

static void
showfile(Directory *dsp, DirEntry *dep)
{
    FILE *inf;

    strcpy(dsp->basep, dep->d_name);

    if ((inf = fopen(dsp->path, "rb")) != NULL) {
        bufcount = fread(iobuf, 1, 1024, inf);
        fclose(inf);
        /*
         * The magic number 13 appears here because code in the
         * print_gif_data() routine assumes the the header is valid.
         */
        if (bufcount >= 13 && strncmp((char*)iobuf, Sgif, 3) == 0) {
            print_gif_data(dep);
        }
    }
}

/*
 * Given the name of a directory, scan its contents and list each
 * appropriate file.
 */

static void
listdir(char *pname)
{
    Directory dir;
    DirEntry de;

    file_count = 0;
    if (verbose) {
        printf("Directory of %s:\n\n", pname);
    }
    if (dir_open(&dir, pname) != 0) {
        FATAL2("\"%s\" is not a valid directory name.\n", pname);
    }
    while (dir_next(&dir, &de) == 0) {
        if (IS_REG(de)) {
            showfile(&dir, &de);
        }
    }
    if (start_column != 0) {
        putchar('\n');
    }
    dir_close(&dir);
    show_count();
}
